-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
--
-- ME Storage Processor
--
--
--                             - If threads ever become real, this thing will fail hard.
-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --

-- TODO: Add fluid and gas support (atleast fluid?, gas might be difficult).

-- Machine processors are defined in me_processor_lib/processors.lua.
-- If anyone adds a machine and gets it working, let me know! The more this program can handle, the better.


-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --


-- FIXME: Hardcoded for now. Should be able to be detected and used if available.
--        Compacting Drawer tester can be a special processor or special device?
--        Learner, drawer and barrel storage could all be special devices, they are inventory controllers.
--        For the compacting tester we need a special processor, it needs a valid transposer.

local DRAWER_ADDRESS = "9619ff36-c6a3-45e2-982f-6dbd0639ecce"
local BARREL_ADDRESS = "5d11f10c-f3b6-4ff3-8878-888e50b4a526"

-- FIXME: Do we really need these still? just ignore unrecognized inventories?
--         ^^^ Yes we do, we want to report really new and unrecognized macchines.
--             But it needs to be more dynamic and east to config then this.
local ignore_inventories = {"opencomputers:rack", "opencomputers:charger", "extrautils2:miner", "enderio:block_aversion_obelisk"}
-- FIXME: same here
--        ^^^ Here we don't, see special devices above and below.
local ignore_transposers = {DRAWER_ADDRESS, BARREL_ADDRESS}


-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --


-- FIXME: Are there better ways to define C-ish debug flags then this?
--THREADING = false
THREADING = true
STICK_KNOWLAGE = true

local DRAWER_MIN = 2500
local DRAWER_TARGET = 5000
local DRAWER_MAX = 10000

local ME_MIN = 1000
local ME_TARGET = 2000
local ME_MAX = 4000

local BARREL_MIN = 100
local BARREL_TARGET = 250
local BARREL_MAX = 10000

local watch_list = {
  ['minecraft:stick|0']   = {min = 1000},
  ['mekanism:sawdust|0']  = {min = 100, target = 200, max = 300}
}

-- These items are processed when above the MAX values.
local recycle_list = {
  "appliedenergistics2:quartz_ore|0", "chisel:marble2|7", "galacticraftcore:basic_block_core|8", "lordcraft:orecrystalb|0",
  "mekanism:oreblock|0", "minecraft:coal_ore|0", "minecraft:diamond_ore|0", "minecraft:gold_ore|0", "minecraft:iron_ore|0",
  "minecraft:lapis_ore|0", "minecraft:redstone_ore|0", "mysticalagriculture:prosperity_ore|0", "mysticalagriculture:inferium_ore|0",
  "netherendingores:ore_other_1|0", "nuclearcraft:ore|4", "nuclearcraft:ore|5", "nuclearcraft:ore|6", "nuclearcraft:ore|7",
  "projectred-exploration:ore|0", "projectred-exploration:ore|1", "projectred-exploration:ore|2", "projectred-exploration:ore|6",
  "thermalfoundation:material|800", "thermalfoundation:material|893", "thermalfoundation:ore|0", "thermalfoundation:ore|1",
  "thermalfoundation:ore|2", "thermalfoundation:ore|3", "thermalfoundation:ore|4", "thermalfoundation:ore|5", "thermalfoundation:ore|6",
  "thermalfoundation:ore|8", "thermalfoundation:ore_fluid|2", "thermalsolars:blocklunarorenether|0", "thermalsolars:blocktitaniumore|0",
  "xreliquary:mob_ingredient|0", "xreliquary:mob_ingredient|2", "xreliquary:mob_ingredient|3", "xreliquary:mob_ingredient|4",
  "xreliquary:mob_ingredient|6", "xreliquary:mob_ingredient|10", "xreliquary:mob_ingredient|11", "xreliquary:mob_ingredient|12"
}

-- Defines one or more processors for an item and how to process that item.
local process_list = {
  ['lootbags:itemlootbag|0']   = {{processor = "loot_opener"}},
  ['lootbags:itemlootbag|1']   = {{processor = "loot_opener"}},
  ['lootbags:itemlootbag|2']   = {{processor = "loot_opener"}},
  ['lootbags:itemlootbag|3']   = {{processor = "loot_opener"}},
  ['lootbags:itemlootbag|4']   = {{processor = "loot_opener"}},
  ['lootbags:itemlootbag|11']  = {{processor = "loot_opener"}},
  ['lootbags:itemlootbag|12']  = {{processor = "loot_opener"}},
  ['lootbags:itemlootbag|13']  = {{processor = "patient_loot_opener"}},
  ['lootbags:itemlootbag|14']  = {{processor = "loot_opener"}},

  ['lordcraft:orecrystalb|0']   = {{processor = "fortune_miner"}},
  ['minecraft:lapis_ore|0']     = {{processor = "fortune_miner"}},
  ['minecraft:redstone_ore|0']  = {{processor = "fortune_miner"}},

  ['mekanism:oreblock|0']               = {{processor = "furnace"}},
  ['minecraft:coal_ore|0']              = {{processor = "furnace"}},
  ['minecraft:diamond_ore|0']           = {{processor = "furnace"}},
  ['minecraft:gold_ore|0']              = {{processor = "furnace"}},
  ['minecraft:iron_ore|0']              = {{processor = "furnace"}},
  ['netherendingores:ore_other_1|0']    = {{processor = "furnace"}},
  ['nuclearcraft:ore|4']                = {{processor = "furnace"}},
  ['nuclearcraft:ore|5']                = {{processor = "furnace"}},
  ['nuclearcraft:ore|6']                = {{processor = "furnace"}},
  ['nuclearcraft:ore|7']                = {{processor = "furnace"}},
  ['projectred-exploration:ore|0']      = {{processor = "furnace"}},
  ['projectred-exploration:ore|1']      = {{processor = "furnace"}},
  ['projectred-exploration:ore|2']      = {{processor = "furnace"}},
  ['projectred-exploration:ore|6']      = {{processor = "furnace"}},
  ['thermalfoundation:ore|0']           = {{processor = "furnace"}},
  ['thermalfoundation:ore|1']           = {{processor = "furnace"}},
  ['thermalfoundation:ore|2']           = {{processor = "furnace"}},
  ['thermalfoundation:ore|3']           = {{processor = "furnace"}},
  ['thermalfoundation:ore|4']           = {{processor = "furnace"}},
  ['thermalfoundation:ore|5']           = {{processor = "furnace"}},
  ['thermalfoundation:ore|6']           = {{processor = "furnace"}},
  ['thermalfoundation:ore|8']           = {{processor = "furnace"}},
  ['thermalsolars:itemlunardust|0']     = {{processor = "furnace"}},
  ['thermalsolars:itemtitaniumdust|0']  = {{processor = "furnace"}},

  ['chisel:marble2|7']                = {{processor = "crafter", recipe = "projectred-exploration:stone|1"}},
  ['thermalfoundation:material|800']  = {{processor = "crafter", recipe = "minecraft:paper|0"}},
  ['xreliquary:mob_ingredient|0']     = {{processor = "crafter", recipe = "minecraft:bone|0"}},
  ['xreliquary:mob_ingredient|2']     = {{processor = "crafter", recipe = "minecraft:spider_eye|0"}},
  ['xreliquary:mob_ingredient|3']     = {{processor = "crafter", recipe = "minecraft:gunpowder|0"}},
  ['xreliquary:mob_ingredient|4']     = {{processor = "crafter", recipe = "minecraft:slime_ball|0"}},
  ['xreliquary:mob_ingredient|6']     = {{processor = "crafter", recipe = "minecraft:gold_nugget|0"}},
  ['xreliquary:mob_ingredient|10']    = {{processor = "crafter", recipe = "minecraft:snowball|0"}},
  ['xreliquary:mob_ingredient|11']    = {{processor = "crafter", recipe = "minecraft:ender_pearl|0"}},
  ['xreliquary:mob_ingredient|12']    = {{processor = "crafter", recipe = "minecraft:dye|0"}},

  ['thermalsolars:blocklunarorenether|0']  = {{processor = "crusher"}},
  ['thermalfoundation:material|893']       = {{processor = "magma_crucible"}},
  ['galacticraftcore:basic_block_core|8']  = {{processor = "double_crusher"}},

  ['appliedenergistics2:quartz_ore|0']      = {{processor = "sag_mill"}, {processor = "pulverizer"}},
  ['mysticalagriculture:inferium_ore|0']    = {{processor = "sag_mill"}, {processor = "pulverizer"}},
  ['mysticalagriculture:prosperity_ore|0']  = {{processor = "sag_mill"}, {processor = "pulverizer"}},
  ['thermalfoundation:ore_fluid|2']         = {{processor = "sag_mill"}, {processor = "pulverizer"}},
  ['thermalsolars:blocktitaniumore|0']      = {{processor = "sag_mill"}, {processor = "pulverizer"}}
}

-- Hard coded recipes.
local recipes = {}
if STICK_KNOWLAGE then
  -- It's always good to know how to make sticks.
  -- To hardcode a recipe, copy one of these lines, adapt it and move it out of the if block.
  recipes['minecraft:planks|0'] = {machine = "crafter", amount = 4, slots = {[5] = "minecraft:log|0"}}
  recipes['minecraft:stick|0']  = {machine = "crafter", amount = 4, slots = {[2] = "minecraft:planks|0", [4] = "minecraft:planks|0"}}
end

-- FIXME: these should be in recipes.dat after being learned or imported.
recipes['minecraft:coal|0']                = {machine = "furnace",        amount = 1, slots = {[1] = "minecraft:coal_ore|0"}}
recipes['projectred-exploration:stone|1']  = {machine = "crafter",        amount = 4, slots = {[1] = "chisel:marble2|7", [2] = "chisel:marble2|7",
                                                                                               [4] = "chisel:marble2|7", [5] = "chisel:marble2|7"}}
recipes['minecraft:gunpowder|0']           = {machine = "crafter",        amount = 6, slots = {[5] = "xreliquary:mob_ingredient|3"}}
recipes['minecraft:slime_ball|0']          = {machine = "crafter",        amount = 6, slots = {[4] = "xreliquary:mob_ingredient|4"}}
recipes['minecraft:paper|0']               = {machine = "crafter",        amount = 6, slots = {[4] = "thermalfoundation:material|800",
                                                                                               [5] = "thermalfoundation:material|800",
                                                                                               [6] = "thermalfoundation:material|800"}}
recipes['minecraft:ender_pearl|0']         = {machine = "crafter",        amount = 3, slots = {[1] = "xreliquary:mob_ingredient|11"}}
recipes['minecraft:gold_nugget|0']         = {machine = "crafter",        amount = 6, slots = {[1] = "xreliquary:mob_ingredient|6",
                                                                                               [2] = "xreliquary:mob_ingredient|6"}}
recipes['minecraft:dye|0']                 = {machine = "crafter",        amount = 6, slots = {[5] = "xreliquary:mob_ingredient|12"}}
recipes['minecraft:bone|0']                = {machine = "crafter",        amount = 5, slots = {[5] = "xreliquary:mob_ingredient|0"}}
recipes['minecraft:spider_eye|0']          = {machine = "crafter",        amount = 2, slots = {[1] = "xreliquary:mob_ingredient|2",
                                                                                               [2] = "xreliquary:mob_ingredient|2"}}
recipes['minecraft:snowball|0']            = {machine = "crafter",        amount = 5, slots = {[5] = "xreliquary:mob_ingredient|10"}}
recipes['potion|0']                        = {machine = "potion_brewer",  amount = 3, slots = {[1] = "minecraft:glass_bottle|0*3",
                                                                                               [2] = "minecraft:nether_wart|0"}}
recipes['mekanism:sawdust|0']              = {machine = "sag_mill",       amount = 1, slots = {[1] = "minecraft:planks|0"}}
recipes['microcontroller|0']               = {machine = "assembler",      amount = 1, slots = {[1] = "opencomputers:material|21",
                                                                                               [17] = "opencomputers:component|0",
                                                                                               [18] = "opencomputers:component|7",
                                                                                               [20] = "opencomputers:storage|0"}}


-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --

-- This code should only call machines.setup() and machine['name'].add_queue + friends.
-- It should only read (and reset) machine['name'].telemetry.

-- FIXME: machines at the moment know about processing, but crafting and processing for machines are the same thing.
--        Abstract that better and change all machines to only process recipes.
--        We can still make on the fly recipes here, machines are dumb, they just do what we tell them to do, and don't
--        care or even need to know what they produce (although, maybe send in a label for them to output on status).
--           ^^^ About label, send them a status prefix like '"Processing " .. item.label' or '"Crafting " .. recipe.item.label'
--               or something along those lines. Just tell the machines to tell us in the future, when we ask, what they are doing.
--               We like to be forgetfull and not keep track of what everybody is up to.

-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --

local computer = require('computer')
local component = require('component')
local serialization = require('serialization')
local shell = require('shell')
local sides = require('sides')

local D = loadfile('me_processor_lib/debug.lua')()

-- FIXME: Change back to require after code has zero bugs.
--local machines = require('me_processor_lib.machines')
--local me_processor_thread = require('me_processor_lib.me_processor_thread')
local machines = loadfile('me_processor_lib/machines.lua')()
local me_processor_thread = loadfile('me_processor_lib/me_processor_thread.lua')()

local me_interface = component.getPrimary("me_interface")

local reserved_items = {}

-- FIXME: reserved_items is a list of items reserved for recursed crafting and processing.
--        Recycling should not touch these and future crafts should refuse if items
--        needed are not in stock when accounting for this list.
--        Recursed crafting threads will remove items from this list when they are done.
--        Recursed crafting will fail if they run low on items directly in stock (within a
--        settable margin of items always kept in stock, depending on the item).


-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --


local function load_recipes()
  if type(next(recipes)) == "nil" and filesystem.exists("recipes.dat") then
    local recipe_data = io.open("recipes.dat", "rb")
    recipes = serialization.unserialize(recipe_data:read("*a"))
    recipe_data:close()
  end
end


-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --


-- FIXME: Possible EMC hack, it can craft things out of nothing, aslong as we have the item.
--        Can we somehow mangle the recipes list to short circuit to EMC if EMC can duplicate the item?
--        Makes autocrafting a whole lot easier.
--          ^^^ Addendum: Or just test here in craft if emc can duplicate it. if it can, it directs to emc.
--                        If it can but not right now, redirect to duplication request, or check if
--                        crafting is 'quick & easy' (whatever that means) and craft if it is.
--                        If it can't, then craft.
--        Another option is to have an EMC database storage system, with 1 of every EMCable item.
--        Can be as simple as me storage cells inside a big crate and just swapping the drive we need into
--        an ME chest.
--        We don't even have to keep configuration state, we can start a thread in the background to walk
--        all drives to catalog their items. Even a full crate of 350 drives should be doable in around 1400 ticks, or under a minute.
--        This gives us a catalog of 22050 items. More then enough for almost all modpacks.
--        Still leaves us with the problem of auto duplicating though. We still need manual duplication link intervention for now.
--        Test if drones can fix that, if not, batteries upon batteries of links?
--        Preferably not connected to ME, but with an RFTools storage scanner.
--        Last idea, do links keep their known inventory? If so, crate with links -> block placer -> block breaker -> Redstone. Done.
--        (If we can ever get a 2nd link that is....)
local function craft(name, damage, amount)
  local item_key = item_to_key(name, damage)
  local recipe = recipes[item_key]

  if not recipe then
    print("Do not know how to craft " .. item_key .. ".")
    return 0
  end

  local machine_name = recipe.machine
  local machine = machines[machine_name]

  if not machine then
    print("Could not find " .. machine_name .. "?")
    os.exit()
  end

  -- FIXME: Would be nice if we could get labels from not yet crafted items.
  --        Cache on the 1st craft? Keep an auto updated database of items?
  --        Item database could potentially be used as an abstract object and tie in with the key to item calls.
  --        Can maybe even be used to store database component item hashes and do NBT stuff.
  --        (Get NBT data via external analyser special device? Maybe with RFTools and some string transferal via signs.)
  return machine:queue_add({recipe = recipe, amount = amount, status_prefix = "Crafting " .. item_key})
end


local function process(name, damage, amount)
  local item_key = item_to_key(name, damage)
  local processors = process_list[item_key]

  if not processors then
    print("Do not know how to process " .. item_key .. ".")
    return 0
  end

  -- FIXME: Take care of situtions where the sag mill is done, but pulverizer still has some item queued.
  --        Recycler does not consider what's on the queues at the moment.
  --        Ties in with reserved_items above.
  for n = 1, #processors do local processor = processors[n]
    local recipe

    if processor.processor == "crafter" then
      local item = key_to_item(processor.recipe)

      -- FIXME: Amount might be wrong here. Look up the recipe amount and work out the proper ratio? Do we even care?
      --        feeding a 1:1 ratio recipe works fine. Depending how we implement the multiplier on for instance the furnace.
      --        Maybe an ignore_multiplier flag or amount_is_input flag... Both ugly and hacky. Find a better way.
      recipe = recipes[processor.recipe]

      if not recipe then
        print("Do not know how to craft " .. item_key .. " to " .. processor.recipe .. ".")
        return 0
      end
    else
      -- Make recipe on the fly.
      recipe = {machine = processor.processor, amount = 1, slots = {[1] = item_key}}
    end

    local machine = machines[processor.processor]

    if not machine then
      print("Could not find " .. processor.processor .. "?")
      os.exit()
    end

    machine:queue_add({recipe = recipe, amount = amount, status_prefix = "Processing " .. item_key})

    -- FIXME: queue_add currently returnes amount queued, but is that useful?
    --        Only info needed is if queueing succeeded.
    --        Here that info can even be ignored (recycler will requeue 60 seconds later).
  end
end


-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --


-- FIXME: From cleanup. Needs to be updated and added as special devices.

local drawers = component.proxy(DRAWER_ADDRESS)
local barrel = component.proxy(BARREL_ADDRESS)
local drawers_side = sides.north
local compactor_side = sides.down
local in_drawers = {}

local function read_drawers()
  in_drawers = {}
  
  for i, stack in pairs(drawers.getAllStacks(drawers_side).getAll()) do
    in_drawers[stack.name .. "|" .. math.floor(stack.damage)] = true
  end
end


local function compactable(item)
  local item_key = item.name .. "|" .. math.floor(item.damage)

  if not in_drawers[item_key] then
    return false
  end

  if is_compactable == nil then
    local compactable_data = io.open("compactable.dat", "rb")

    if compactable_data ~= nil then
      is_compactable = serialization.unserialize(compactable_data:read("*a"))
      compactable_data:close()
    end

    if is_compactable == nil then
      is_compactable = {}
    end
  end

  if is_compactable[item_key] ~= nil then
    return is_compactable[item_key]
  end

  if item.size < 9 then
    return false
  end

  local drawer_slot = 0
  for i, stack in pairs(drawers.getAllStacks(drawers_side).getAll()) do
    if (stack.name == item.name) and (stack.damage == item.damage) then
      drawer_slot = i
      break
    end
  end
  
  if (drawer_slot == 0) then
    print("Could not find " .. item_key .. " in drawers?")
    
    return false
  end

  drawers.transferItem(drawers_side, compactor_side, 9, drawer_slot)
  local compactor_slot = 0

  for i, stack in pairs(drawers.getAllStacks(compactor_side).getAll()) do
    if (stack.size == 9) then
      compactor_slot = i
    elseif (stack.size > 0) and (stack.size < 9) then
      is_compactable[item_key] = true
    end
  end

  if (compactor_slot == 0) then
    print("Could not find " .. item_key .. " in compactor? Flushing Compactor...")
    drawers.transferItem(compactor_side, drawers_side, 9*9*9)
  else
    drawers.transferItem(compactor_side, drawers_side, 9, compactor_slot)
  end

  if is_compactable[item_key] == nil then
    is_compactable[item_key] = false
  end

  local compactable_data = io.open("compactable.dat", "wb")
  compactable_data:write(serialization.serialize(is_compactable))
  compactable_data:close()

  return is_compactable[item_key]
end


-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --


-- FIXME: Mainly a straight copy from cleanup. Probably need to clean this up a bit.
local function recycler()
  while true do
    print("Starting cleanup...")

    read_drawers()

    local filter = {hasTag = false, maxDamage = 0}  -- Preselection.
    local items = me_interface.getItemsInNetwork(filter)

    if not items then
      items = {n = 0}
    end

    print("Selected " .. items.n .. " types of items to consider for recycling...")
    
    for n = 1, items.n do
      item = items[n]

      if item.maxSize > 1 then  -- FIXME hack to filter out barrel items. Do it properly with a barrel check.
        local item_key = item_to_key(item)
        local emc_value = machines.emc:value(item_key) or 0

        -- FIXME: machine.compactable:is_compactable(item_key)?
        --        And of course machine.drawers:in_drawers(item_key) or something.
        --        Abstract everything out of here to the machines interface.
        --        Can be a special class of machine, like analyser. Also good or the recipe learner.
        if emc_value > 0 then
          if in_drawers[item_key] and not compactable(item) and item.size > DRAWER_MAX then
            machines.emc:queue_add({recipe = {amount = 1, slots = {[1] = item_key}}, amount = item.size - DRAWER_MAX, status_prefix = "Transmutating " .. item_key})
          elseif in_drawers[item_key] and not compactable(item) and item.size < DRAWER_MIN then
--            machines.emc:queue_add({duplicate = item_key, amount = item.size - DRAWER_TARGET})
          elseif not in_drawers[item_key] and item.size > ME_MAX then
            machines.emc:queue_add({recipe = {amount = 1, slots = {[1] = item_key}}, amount = item.size - ME_MAX, status_prefix = "Transmutating " .. item_key})
          elseif (not in_drawers[item_key]) and (item.size < ME_MIN) then
--            machines.emc:queue_add({duplicate = item_key, amount = item.size - ME_TARGET})
          end
        else
          -- FIXME: Stuff in drawers which oare on the recycle list should go to DRAWER_MAX.
          for i = 1, #recycle_list do
            if item_key == recycle_list[i] and item.size > ME_MAX then
              process(item.name, item.damage, item.size - ME_MAX)
              break
            end
          end
        end
      end
    end

    -- Lootbags.
    local lootbags = me_interface.getItemsInNetwork({name = "lootbags:itemlootbag"})
    if lootbags then
      print("Recycling Loot Bags...")
      
      for n = 1, lootbags.n do
        process(lootbags[n].name, lootbags[n].damage, lootbags[n].size)
      end
    else
      print("No Loot Bags to recycling.")
    end

--FIXME: when queues have limits and check for items already on queue, change back to 60.
    print("Cleanup done... Sleeping 120 seconds.")

    os.sleep(120)
    print()
  end
end


-- Should run every second, if main() is doing it's job.
function update_report()
  -- FIXME: Read telemetry, print nice status. 
--  print("Status: ???")
end


local function main(args)
  if args[1] and args[1] == "-t" then
    THREADING = false
  end

  machines = machines.setup(ignore_inventories, ignore_transposers)
  load_recipes()
  
  local next_update = computer.uptime()

  me_processor_thread.create(recycler)

  while true do
    local now = computer.uptime()
    
    if now >= next_update then
      if (next_update - now) > 0.5 then
        print("Hung for " .. (now - next_update) .." seconds?")
      end
      
      next_update = now + 1

      update_report()
    elseif now < (next_update - 1) then
      print("Clock skew? (" .. (next_update - 1) .. " -> " .. now .. ").")

      next_update = now
    end

    me_processor_thread.sleep(next_update - computer.uptime())
  end
end

-- Run in a thread so we can kill the whole process from any thread if needed.
me_processor_thread.run_main_thread(main, shell.parse(...))
