--local computer = require('computer')
--local component = require('component')
local filesystem = require('filesystem')
--local term = require('term')
local shell = require('shell')
--local sides = require('sides')
local serialization = require('serialization')

local me_processor_thread = require('me_processor_lib.me_processor_thread')
--local me_processor_thread = loadfile('me_processor_lib.me_processor_thread')()

local machines = {}
local D = loadfile('me_processor_lib/debug.lua')()

--[[

require('me_processor.library')

--]]


-- FIXME: Change back to require after code has zero bugs.
------local interface_factory = require('me_processor_lib.interface_factory')
--local transposer_factory = require('me_processor_lib.transposer_factory')
--local processors = require('me_processor_lib.processors')

----local interface_factory = loadfile('me_processor_lib/interface_factory.lua')()
local transposer_factory = loadfile('me_processor_lib/transposer_factory.lua')()
local processors = loadfile('me_processor_lib/processors.lua')()




--[[







local machines_known = {}
local transposer = {}
local emc = {}
local threads = {}
local processor_init = {}

local TE_MACHINE_ID = "thermalexpansion:machine"
local LOOT_OPENER_ID = "lootbags:loot_opener"
local LOOT_OPENER_ADDRESS = "449956f9-d9e5-40d2-a0db-16175334292e"
local RECIPE_LEARNER_ADDRESS = "449956f9-d9e5-40d2-a0db-16175334292e"





--]]


-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --


local function thread_run(machine, processor_index)
  if #machine.queue == 0 then
    machine.running = false
    return
  end
  
  -- Yield to main thread for more queueing.
  os.sleep(0)

  print("Starting " .. machine.label .. " thread.")
  machine.thread(machine, processor_index)
  print(machine.label .. " thread done.")

  machine.running = false
end


local function start_thread(machine, processor_index)
  if not machine.running then
    machine.running = true
    if THREADING then
      me_processor_thread.create(thread_run, machine, processor_index)
    else
      thread_run(machine, processor_index)
    end
  end
end


-- Queueing:
-- each processor takes an amount of the topqueue -> amount = (amount_left / #processors) and at least maxSize (64).
-- This ensures automatic and maximal parallelization for both big and small (<64) amounts

function queue_add(self, request)
  table.insert(self.queue, request)

  start_thread(self, 1)
end


-- FIXME: TBD
function queue_cancel(self, request)
end


-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --




-- Adds to processor base to turn it into a machine.
local function new_machine(machine_name)
  local machine = processors.processor_base(machine_name)

  machine.processors = {}
  machine.running = false
  machine.queue = {}
  machine.queue_add = queue_add
  machine.queue_remove = queue_remove

  return machine
end


-- Turns configuration into valid machines.
local function validate_configuration(configuration)
  machines = {}
  
  if not configuration or not configuration.known_processors or not configuration.transposers then
    print("Configuration invalid.")
    return false
  end

  local known_processors = configuration.known_processors
  local transposer_configuration = configuration.transposers

  for processor_name, processor_label in pairs(processors.all()) do
    if not known_processors[processor_name] or known_processors[processor_name] ~= processor_label then
      -- Processor added in source.
      print("New processor type found (" .. processor_label .. ").")
    end
  end

  -- Turns configuration into valid transposers.
  local transposers = transposer_factory.validate_transposer_configuration(transposer_configuration)

  if not transposers then
    print("Transposer validation failed.")
    return false
  end

  if transposers == {} then
    print("WARNING: No transposers found.")

    -- All perfectly fine and dandy though.
    return machines
  end

  -- Add the machines.
  for transposer_address, transposer in pairs(transposers) do
    io.write("Setting up transposer " .. transposer.address .. " with " .. #transposer.interfaces .. " interface" .. (#transposer.interfaces > 1 and "s" or ""))

    local i = 0
    for n, processor in pairs(transposer.processors) do
      if n == 1 then
        io.write(": " .. processor.label)
      else
        io.write(", " .. processor.label)
      end
      
      if not machines[processor.name] then
        machines[processor.name] = new_machine(processor.name)
      end

      table.insert(machines[processor.name].processors, {transposer = transposer, side = processor.side})
    end

    print(".")
  end

  if machines == {} then
    -- Still all completely okey.
    print("WARNING: No machines found.")
  end

  for machine_name, machine in pairs(machines) do
    processors.initialize_machine(machine)
  end

  return true
end

local function setup(ignore_inventories, ignore_transposers)
  machines = {}

  if filesystem.exists(shell.getWorkingDirectory() .. "/configuration.dat") then
    print("Found configuration.dat. Loading configuration from disk.")

    local configuration_data = io.open("configuration.dat", "rb")
    local configuration = serialization.unserialize(configuration_data:read("*a"))
    configuration_data:close()

    if validate_configuration(configuration) then
      print("Configuration validated.")
      print()

      return machines
    end

    print("Configuration update detected.")
  end

  local configuration = {
    known_processors = processors.all(),
    transposers = transposer_factory.detect_transposer_configuration(ignore_inventories, ignore_transposers)
  }

  if not validate_configuration(configuration) then
    print("Internal error - validate_configuration() failed to validate detected configuration.")
    print("Cannot continue.")
    os.exit()
  end

  local configuration_data = io.open("configuration.dat", "wb")
  configuration_data:write(serialization.serialize(configuration))
  configuration_data:close()

  print("Setup complete.")
  print()

  return machines
end

return {setup = setup}
