local computer = require('computer')

local emc = {}

-- FIXME: TBD correctly. Worked in cleanup.lua, not here.
local function handle_emc_duplicate(machine, item_key, amount)
  local item = key_to_item(item_key)
  item = interface.getItemsInNetwork({name = item.name, damage = item.damage})

  if item[1] then
    item = item[1]
  else
    print("Could not find " .. item_key .. " on the network?")
    return 0
  end

  local max_size = item.maxSize
  request_item(machine, item_key, 1, (max_size > amount and amount or max_size))

  local moved = 0

  repeat
    local current_amount = amount
    
    if current_amount > max_size then
      current_amount = max_size
    end

    if not wait_for_request(machine, item_key, 1, current_move, 1) then
      print("Timeout while waiting for " .. current_amount .. " of " .. item_key .. ".")
      return moved
    else
      moved = moved + move_items(machine, machine.interface_side, 1, machine.processor_side, 1, current_amount)
    end
  until moved >= amount

  cancel_request(machine, 1)
end


local function update_duplication()
  in_emc = {}
  for i, item in pairs(transposer.getAllStacks(emc_side).getAll()) do
    if item.size > 0 then
      in_emc[item.name .. "|" .. math.floor(item.damage)] = i
    end
  end

  duplication_room = 54
  in_duplication = {}
  for i, item in pairs(transposer.getAllStacks(duplication_side).getAll()) do
    if item.size > 0 then
      in_duplication[item.name .. "|" .. math.floor(item.damage)] = true
      duplication_room = duplication_room - 1
    end
  end

  print("Duplication room: " .. duplication_room)
end


local function emc_duplicate(self, item_key, amount)
  if in_emc[item_key] then
    print("Getting " .. math.floor(amount) .. " " .. item.label .. " from EMC.")

    local moved = transposer.transferItem(emc_side, duplication_side, 15, in_emc[item_key])
    
    while moved < amount do
      local current_move = amount - moved
      
      if current_move > item.maxSize then
        current_move = item.maxSize
      end

      local current_moved = transposer.transferItem(emc_side, interface_side, current_move, in_emc[item_key])
      
      if current_moved == 0 then
        break
      end
      
      moved = moved + current_moved
    end

    return
  end

  if in_duplication[item_key] or duplication_room == 0 then
    return
  end

  print("Requesting " .. item.label .. " to be duplicated.")

  database.clear(self.database_slot)
  me_interface.store({name = item.name, damage = item.damage}, database.address, self.database_slot)
  me_interface.setInterfaceConfiguration(1, database.address, self.database_slot, 1)

  while not transposer.compareStackToDatabase(interface_side, 1, database.address, self.database_slot, true) do
    os.sleep(0.05)
  end

  transposer.transferItem(interface_side, duplication_side, 1, 1)
  me_interface.setInterfaceConfiguration(1)

  duplication_room = duplication_room - 1
  in_duplication[item_key] = true
end

function processor_thread.emc(machine, processor_index)
  repeat
    request = machine.queue[1]

    -- Reuse the generic handler for processing.
    if request.recipe then
      handle_generic_recipe(machine, machine.processors[processor_index], request.recipe, request.amount, request.status_prefix)
    elseif request.duplicate then
      handle_emc_duplicate(machine, machine.queue[1].duplicate, machine.queue[1].amount)
    else
      print("Queued request invalid for EMC. Ignored request.")
    end
    table.remove(machine.queue, 1)
  until #machine.queue == 0
end


local function load_emc()
  local start_time = computer.uptime()
  print("Reading EMC Table...")

  emc = io.open("emc", "rb")
  local emc_data = emc:read("*a")
  emc:close()

  emc = {}
  local last_sleep = computer.uptime()
  for line in string.gmatch(emc_data, "([^\n]+)") do
    if line == "" then
      break
    end

    local part = string.gmatch(line, "([^\t]+)")
    local item_key = part()
    local emc_value = part()

    emc[item_key] = tonumber(emc_value)

    if computer.uptime() - last_sleep > 1 then
      os.sleep(0)
      last_sleep = computer.uptime()
    end  
  end

  print("EMC table loaded in " .. string.format("%0.2f", (computer.uptime() - start_time)) .. "s.")
  print()
end

local function value(self, item_key)
  if emc[item_key] then
    return emc[item_key]
  end
  
  return 0
end

function processor_init.emc(machine)
  load_emc()
--  print("EMC not loaded.")

  machine.value = value
end
