-- FIXME: Check if we auto import or not?
--        We can test this the first craft with a processor and see if it auto imports it's results.
--        Warn if a processor doesn't?
--        Also, with 9 slot processors (crafters) returning items via the interface is tricky and slower.
-- FIXME: Add back Analog crafter as a basic crafting processor.
--        Check if Void whatyamacallit would work too? (Void Lordcraft crafter).
local processors = {
  ['emc']                    = {id = "projectex:compressed_refined_link", label = "EMC Link", thread = "emc", init="emc", stack_time = 1},
  ['assembler']              = {id = "opencomputers:assembler", label = "Electronics Assembler", thread = "generic", craft = "assembler"},
  ['pulverizer']             = {id = "thermalexpansion:machine", label = "Pulverizer", thread = "generic"},
  ['compactor']              = {id = "thermalexpansion:machine", label = "Compactor", thread = "generic"},
  ['induction_smelter']      = {id = "thermalexpansion:machine", label = "Induction Smelter", thread = "generic"},
  ['insolator']              = {id = "thermalexpansion:machine", label = "Phytogenic Insolator", thread = "generic"},
  ['magma_crucible']         = {id = "thermalexpansion:machine", label = "Magma Crucible", thread = "generic"},
  ['sag_mill']               = {id = "enderio:block_enhanced_sag_mill", label = "Sag Mill", thread = "generic"},
  ['furnace']                = {id = "furnaceoverhaul:zenith_furnace", label = "Zenith Furnace", thread = "generic", stack_time = 60, multiplier = 2},
  ['alloy_smelter']          = {id = "enderio:block_enhanced_alloy_smelter", label = "Alloy Smelter", thread = "generic"},
  ['loot_opener']            = {id = "lootbags:loot_opener", label = "Bag Opener", thread = "processing", init="loot_opener"},
  ['patient_loot_opener']    = {id = "lootbags:loot_opener", label = "Patient Bag Opener", thread = "patient_loot_opener"},
  ['crafter']                = {id = "opencomputers:robot", label = "Crafting Robot", thread = "generic", init = "crafter", craft = "crafter",
                                      slot_map = {[1] = 1, [2] = 2, [3] = 3, [4] = 5, [5] = 6, [6] = 7, [7] = 9, [8] = 10, [9] = 11}},
                                      -- 'crafter' thread is for this robot, 'crafting' thread is generic.
  ['hellfire_forge']         = {id = "bloodmagic:soul_forge", label = "Hellfire Forge", thread = "generic"},
  ['enchantment_extractor']  = {id = "industrialforegoing:enchantment_extractor", label = "Enchantment Extractor", thread = "generic"},
  ['potion_brewer']          = {id = "industrialforegoing:potion_enervator", label = "Potion Brewer", thread = "potion_brewer",
                                      slot_map = {[1] = 7, [2] = 8, [3] = 9, [4] = 10, [5] = 11, [6] = 12}},
  ['fortune_miner']          = {id = "extrautils2:user", label = "Fortune III Miner", thread = "generic"},
  ['crusher']                = {id = "extrautils2:machine", label = "Crusher", thread = "generic"},
  ['double_crusher']         = {id = "actuallyadditions:block_grinder_double", label = "Double Crusher", thread = "generic"},
  ['enricher']               = {id = "mekanism:machineblock", label = "Elite Enriching Factory", thread = "generic"},
  ['compressor']             = {id = "galacticraftcore:machine2", label = "Electric Compressor", thread = "generic"}
}

-- FIXME: I really want these local, not global. Find out how.
processor_init = {}
processor_thread = {}
processor_craft = {}

require('me_processor_lib.processor_generic')
require('me_processor_lib.processor_assembler')
require('me_processor_lib.processor_crafter')
require('me_processor_lib.processor_emc')
require('me_processor_lib.processor_loot_opener')
require('me_processor_lib.processor_patient_loot_opener')
require('me_processor_lib.processor_potion_brewer')

local DEFAULT_STACK_TIME = 256

local TE_MACHINE_ID = "thermalexpansion:machine"
local LOOT_OPENER_ID = "lootbags:loot_opener"

-- Should be not here.
local LOOT_OPENER_ADDRESS = "d88b0919-2418-46e0-857c-391da55d7f18"
local RECIPE_LEARNER_ADDRESS = "449956f9-d9e5-40d2-a0db-16175334292e"


-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --


local function all()
  processor_list = {}
  
  for processor_name, processor in pairs(processors) do
    table.insert(processor_list, processor_name)
  end
  
  return processor_list
end


local function identify_processor(transposer, side)
  local inventory_name = transposer.getInventoryName(side)
  if not inventory_name then
    return {}
  end

  if inventory_name == TE_MACHINE_ID then
    local inventories = transposer.getInventorySize(side)

    local tanks = 0
    while pcall(transposer.getTankCapacity, side, tanks + 1) do
      tanks = tanks + 1
    end

    local processor

    if inventories == 0 and tanks == 1 then
      -- Ignore the Fluid Transposers.
      return
    elseif inventories == 1 and tanks == 1 then
      processor = "magma_crucible"
    elseif inventories == 2 and tanks == 0 then
      processor = "compactor"
    elseif inventories == 3 and tanks == 0 then
      processor = "pulverizer"
    elseif inventories == 4 and tanks == 0 then
      processor = "induction_smelter"
    elseif inventories == 4 and tanks == 1 then
      processor = "insolator"
    end

    if processor then
      return {name = processor, label = processors[processor].label, side = side}
    end

    print("Unknown TE machine (inventories: " .. math.floor(inventories) .. ", tanks: " .. math.floor(tanks) .. ").")

    return
  elseif inventory_name == LOOT_OPENER_ID then
    -- FIXME: Hardcoded address for now. Transposer on this address has a barrel, can we check on that? (Without a 6 side search?)
    --        Also, side here should be the barrel side to get rid of the other lootbag hack.
    --        This Information is only available just before transposer_setup, so add a hack there to get rid of these 2 and the hardcoded address?
    --        Possibly also the place to detect the compacting tester. So a general hack_hook there for custom transposer hacks?
    --        We want Anti Barrels as an option for all machines, so can't use it as an identifier for the loot bag machine.
    if transposer.address == LOOT_OPENER_ADDRESS then
      return {name = "loot_opener", label = processors.loot_opener.label, side = side}
    else
      return {name = "patient_loot_opener", label = processors.patient_loot_opener.label, side = side}
    end
  end

  for name, processor in pairs(processors) do
    if inventory_name == processor.id then
      return {name = name, label = processor.label, side = side}
    end
  end

  print("Unknown inventory: " .. inventory_name)
end


function processor_base(processor_name)
  local processor_base = {
    name = processor_name,
    label = processors[processor_name].label,
    thread = processor_thread[processors[processor_name].thread]
  }

  if processors[processor_name].craft then
    processor_base.craft = processor_craft[processors[processor_name].craft]
  end

  if processors[processor_name].stack_time then
    processor_base.stack_time = processors[processor_name].stack_time
  else
    processor_base.stack_time = DEFAULT_STACK_TIME
  end

  processor_base.slot_map = {}
  -- Max 5 interfaces possible on 1 transposer.
  for i = 1, (9 * 5) do
    processor_base.slot_map[i] = i
  end

  if processors[processor_name].slot_map then
    for slot, target_slot in pairs(processors[processor_name].slot_map) do
      processor_base.slot_map[slot] = target_slot
    end
  end

  return processor_base
end


-- This function takes full machines.
-- Needs to be here to call the processor init after the machine is completed by machines.
-- Little bit of a layer break, but whatever.
-- Could just move it into machines I guess, but I don't want the threading functions exposed there.
function initialize_machine(machine)
  if processors[machine.name].init then
    processor_init[processors[machine.name].init](machine)
  end
end


return {
  all = all,
  identify_processor = identify_processor,
  processor_base = processor_base,
  initialize_machine = initialize_machine
}
